def simulate_single_mutants(model, list_nodes=[], sign="BOTH", cmaboss=False):
    """
        Simulates a batch of single mutants and return an array of results

        :param model: the model on which to perform the simulations
        :param list_nodes: the node(s) which are to be mutated (default: all nodes)
        :param sign: which mutations to perform. "ON", "OFF", or "BOTH" (default)
    """ 
    
    if len(list_nodes) == 0:
        list_nodes = list(model.network.keys())
    
    list_single_mutants = []
    if (sign == "BOTH" or sign == "ON"):
        list_single_mutants += [(node, "ON") for node in list_nodes]
    if (sign == "BOTH" or sign == "OFF"):
        list_single_mutants += [(node, "OFF") for node in list_nodes]
    
    res = {}
    for single_mutant in list_single_mutants:
        t_model = model.copy()
        t_model.mutate(*single_mutant)
        res.update({single_mutant: t_model.run(cmaboss=cmaboss)})
        
    return res
    
    
def simulate_double_mutants(model, list_nodes, sign="BOTH", cmaboss=False):
    """
        Simulates a batch of double mutants and return an array of results

        :param model: the model on which to perform the simulations
        :param list_nodes: the node(s) which are to be mutated (default: all nodes)
        :param sign: which mutations to perform. "ON", "OFF", or "BOTH" (default)
    """ 
    
    if len(list_nodes) == 0:
        list_nodes = list(model.network.keys())
    
    list_single_mutants = []
    if (sign == "BOTH" or sign == "ON"):
        list_single_mutants += [(node, "ON") for node in list_nodes]
    if (sign == "BOTH" or sign == "OFF"):
        list_single_mutants += [(node, "OFF") for node in list_nodes]
    
    list_double_mutants = [(a, b) for idx, a in enumerate(list_single_mutants) for b in list_single_mutants[idx + 1:] if a[0] != b[0]]

    res = {}
    for double_mutant in list_double_mutants:
        t_model = model.copy()
        t_model.mutate(*(double_mutant[0]))
        t_model.mutate(*(double_mutant[1]))
        res.update({double_mutant: t_model.run(cmaboss=cmaboss)})
        
    return res
    
def filter_sensitivity(results, state=None, node=None, minimum=None, maximum=None):
    """
        Filter a list of results by state of nodes value

        :param results: the list of results to filter
        :param state: the state on which to apply the filter (default None)
        :param node: the state on which to apply the filter (default None)
        :param minumum: the minimal value of the node (default None)
        :param maximum: the maximal value of the node (default None)

        Example : 

        Filtering results showing more than 50% for Proliferation node
        >>> res_ensemble = filter_sensitivity(results, node='Proliferation', maximum=0.5)

        Filtering results showing more than 10% for Apoptosis -- NonACD state
        >>> res_ensemble = filter_sensitivity(results, state='Apoptosis -- NonACD', minimum=0.1)
    """ 
    ret_res = {}
    for (mutant, res) in results.items():
        
        if state is not None:
            t_res = res.get_last_states_probtraj()
            if state in t_res.columns:
                if minimum is not None and maximum is not None:
                    if t_res[state].values[0] >= minimum and t_res[state].values[0] <= maximum:
                        ret_res.update({mutant: res})
                elif minimum is not None and t_res[state].values[0] >= minimum:
                    ret_res.update({mutant: res})
                elif maximum is not None and t_res[state].values[0] <= maximum:
                    ret_res.update({mutant: res})
            elif maximum is not None and minimum is None:
                ret_res.update({mutant:res})

        elif node is not None:
            
            t_res = res.get_last_nodes_probtraj()

            if node in t_res.columns:
                if minimum is not None and maximum is not None:
                    if t_res[node].values[0] >= minimum and t_res[node].values[0] <= maximum:
                        ret_res.update({mutant: res})
                elif minimum is not None and t_res[node].values[0] >= minimum:
                    ret_res.update({mutant: res})
                elif maximum is not None and t_res[node].values[0] <= maximum:
                    ret_res.update({mutant: res})
            elif maximum is not None and minimum is None:
                ret_res.update({mutant: res})

    return ret_res



from .network import *
from .simulation import *
from .result import *
from .results import *
from .gsparser import load, loadBNet, loadSBML
from .server import MaBoSSClient
from .upp import UpdatePopulation
from .ensemble import EnsembleResult, Ensemble
from .pop import PopSimulation
import platform 
import maboss.pipelines

from colomoto_jupyter import IN_IPYTHON
import colomoto.setup_helper

if IN_IPYTHON:
    from colomoto_jupyter import jupyter_setup
    from .widgets import *

    import matplotlib.pyplot as plt
    plt.ion()

    jupyter_setup("mymodule",
        label="MaBoSS",
        color="green", # for menu and toolbar
        menu=menu,
        toolbar=toolbar,
        js_api=js_api)


if platform.system() == "Windows":
    bin_name = "MaBoSS.exe"
else:
    bin_name = "MaBoSS"
if shutil.which(bin_name) is None:

    if platform.system() == "Windows":
        bin_path = os.path.join(os.getenv("APPDATA"), "maboss", "bin")
    else:
        bin_path = os.path.join(os.path.expanduser("~"), ".local", "share", "maboss", "bin")

    if os.path.exists(bin_path):
        if platform.system() == "Windows":
            os.environ["PATH"] = "%s;%s" % (bin_path, os.environ["PATH"])
        else:
            os.environ["PATH"] = "%s:%s" % (bin_path, os.environ["PATH"])



from __future__ import print_function
import sys
from .results import UpdatePopulationResults
from .cmaboss_results import CMaBoSSUpdatePopulationResults

class UpdatePopulation:
    """
        .. py:class:: construct a simulation for Update Population MaBoSS.

        :param model: MaBoSS model
        :param uppfile: upp file, default to None 
        :param previous_run: previous run to start from, default to None 
        :param nodes_init: dict in the form { "node1" : TrueProb1, "node2" : TrueProb2, ...} with nodes to be initialised to a specific value at the start of the simulation. These values override the previous run probabilities for the specified nodes, default to None 
        :param verbose: boolean to activate verbose mode, default to False
        
        """
        
    def __init__(self, model, uppfile=None, previous_run=None, 
                 nodes_init=None, verbose=False):

        self.model = model
        self.uppfile = uppfile
        self.previous_run = previous_run
        self.nodes_init = nodes_init

        self.time_step = float(model.param['max_time'])
        self.time_shift = 0.0
        self.base_ratio = 1.0

        self.node_list = list(model.network.keys())
        self.division_node = ""
        self.death_node = ""

        self.update_var = {}
        self.nodes_formula = {}
        self.pop_ratio = 1.0
        self.step_number = -1

        self.verbose = verbose

        if previous_run:
            # Chain the results!!
            prev_pop_ratios = previous_run.get_population_ratios()
            self.time_shift = prev_pop_ratios.last_valid_index()
            self.base_ratio = prev_pop_ratios.iloc[-1]
            self.model = model.copy()

        if self.uppfile is not None:
            if "://" in self.uppfile:
                from colomoto_jupyter.io import ensure_localfile
                self.uppfile = ensure_localfile(self.uppfile)

            self._readUppFile()

    def run(self, workdir=None, overwrite=None, verbose=False, host=None, port=7777, cmaboss=False, only_final_state=False):
        """
        .. py:method:: Runs the simulation

        :param workdir: (optional) Working directory in which to save result files
        :param overwrite: (optional) Boolean to indicate if you want to overwrite existing results in the working directory
        :param verbose: (optional) Boolean to indicate if you want debugging information
        :param host: (optional) Host to use when simulating on a MaBoSS server
        :param port: (optional) Port to use when simulating on a MaBoSS server
        :return: The Update Population results object.
        """
        
        if cmaboss:
            return CMaBoSSUpdatePopulationResults(self, verbose, workdir, overwrite, self.previous_run, nodes_init=self.nodes_init, only_final_state=only_final_state)
        else:
            return UpdatePopulationResults(self, verbose, workdir, overwrite, self.previous_run, host=host, port=port, nodes_init=self.nodes_init)

    def _readUppFile(self):

        try:
            with open(self.uppfile, 'r') as UPP:
                for line in UPP.readlines():
                    
                    if line.startswith("death"):
                        
                        if self.death_node != "":
                            print("Multiple definition of death node", file=sys.stderr)
                            exit()

                        self.death_node = line.split("=")[1]
                        self.death_node = self.death_node.replace(";", "").strip()
                       
                        if self.verbose:
                            print("Death node : %s" % self.death_node)

                    if line.startswith("division"):
                        
                        if self.division_node != "":
                            print("Multiple definition of division node", file=sys.stderr)
                            exit()

                        self.division_node = line.split("=")[1]
                        self.division_node = self.division_node.replace(";", "").strip()
                        
                        if self.verbose:
                            print("Division node : %s" % self.division_node)

                    if line.startswith("steps"):
                        
                        if self.step_number != -1:
                            print("Multiple definition of step number", file=sys.stderr)
                            exit()

                        self.step_number = line.split("=")[1]
                        self.step_number = int(self.step_number.replace(";", "").strip())
                        
                        if self.verbose:
                            print("Number of steps : %s" % self.step_number)
            
                    if line.startswith("$"):

                        (varName, value) = line.split("u=", 1)
                        varName = varName.strip()
                        if varName in self.update_var.keys():
                            print("Multiple definitions of %s" % varName)
                            exit()

                        value = value.replace(";", "").strip()
                        self.update_var.update({varName: value})
                        
                        if self.verbose:
                            print("Var %s updated by value %s" % (varName, value))
                            
                    if line.startswith("@"):

                        (nodeName, value) = line.split("u=", 1)
                        nodeName = nodeName.strip()[1:] # remove starting @
                        if nodeName in self.nodes_formula.keys():
                            print("Multiple formula definitions for node %s" % nodeName)
                            exit()

                        value = value.replace(";", "").strip()
                        self.nodes_formula.update({nodeName: value})
                        if self.verbose:
                            print("Node %s has formula %s" % (nodeName, value))

        except FileNotFoundError:
            print("Cannot find .upp file", file=sys.stderr)
            exit()

    def setStepNumber(self, step_number):
        """
            .. py:method:: Modifies the number of step of the simulation

            :param step_number: The number of steps of the simulation
        """
        self.step_number = step_number

    def setDeathNode(self, death_node):
        """
            .. py:method:: Modifies the identifier of the node used to represent cellular death

            :param death_node: The identifier of the node to be used to represent death
        """
        self.death_node = death_node

    def setDivisionNode(self, division_node):
        """
            .. py:method:: Modifies the identifier of the node used to represent cellular division
            
            :param division_node: The identifier of the node to be used to represent division
        """
        self.division_node = division_node

    def setExternalVariable(self, name, formula, overwrite=False):
        """
            .. py:method:: Creates a rule to update the external variable at each step
            
            :param name: The identifier of the variable to be updated
            :param formula: The formula for the variable's update
            :param overwrite: (optional) Overwrite the rule if one already exists for the given name
        """
        if name in self.update_var.keys() and not overwrite:
            print("External variable %s already exists !" % name, file=sys.stderr)
            return

        self.update_var.update({name: formula})

    def setNodeFormula(self, node, formula, overwrite=False):
        if node in self.nodes_formula.keys() and not overwrite:
            print("Formula for node %s already exists !" % node, file=sys.stderr)
            return

        self.nodes_formula.update({node: formula})



"""
Class that contains the cMaBoSS Population simulation.
"""
# from .sbmlcmabossresult import SBMLCMaBoSSResult
# from sys import stdout
import os, sys
from .popresult import PopMaBoSSResult
from .cmabosspopresult import cMaBoSSPopMaBoSSResult
from .storedpopresult import StoredPopResult

class PopSimulation(object):

    def __init__(self, network=None, cfgs=None, maboss_model=None, cmaboss=None, cmaboss_sim=None):

        if maboss_model is not None:
            self.cmaboss = maboss_model.cmaboss
            self.cmaboss_sim = self.cmaboss.PopMaBoSSSim(network_str=maboss_model.str_bnd(), config_str=maboss_model.str_cfg())
        
        elif cmaboss is not None and cmaboss_sim is not None:
            self.cmaboss = cmaboss
            self.cmaboss_sim = cmaboss_sim
        else:
            nb_nodes = self.count_nodes(network)
            self.cmaboss = self.get_cmaboss(nb_nodes)

            if cfgs is None:
                self.cmaboss_sim = self.cmaboss.PopMaBoSSSim(network)
            else:
                if isinstance(cfgs, str):
                    self.cmaboss_sim = self.cmaboss.PopMaBoSSSim(network, config=cfgs)
                elif isinstance(cfgs, list):
                    self.cmaboss_sim = self.cmaboss.PopMaBoSSSim(network, configs=cfgs)

        self.param = self.cmaboss_sim.param
        self.network = self.cmaboss_sim.network
        
    def count_nodes(self, network):
        
        res = 0
        with open(network, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if line.lower().startswith("node"):
                    res += 1
        return res
    
    def get_cmaboss(self, nb_nodes):

        assert nb_nodes <= 1024, "Models with more than 1024 nodes are not compatible with this version of MaBoSS"

        if nb_nodes <= 64:
            return __import__("cmaboss")
        elif nb_nodes <= 128:
            return __import__("cmaboss_128n")
        elif nb_nodes <= 256:
            return __import__("cmaboss_256n")
        elif nb_nodes <= 512:
            return __import__("cmaboss_512n")
        else:
            return __import__("cmaboss_1024n")

    def update_parameters(self, **kwargs):
        self.cmaboss_sim.update_parameters(**kwargs)

    def set_custom_pop_output(self, formula):
        self.cmaboss_sim.set_custom_pop_output(formula)
    
    def copy(self):
        return PopSimulation(cmaboss=self.cmaboss, cmaboss_sim=self.cmaboss_sim.copy())
        
    # def get_nodes(self):
    #     return self.cmaboss_sim.get_nodes()

    def run(self, workdir=None, prefix="res", overwrite=False, hexfloat=True, cmaboss=True):
        if workdir is None or not os.path.exists(os.path.join(workdir, "%s_run.txt" % prefix)) or overwrite:
            return cMaBoSSPopMaBoSSResult(self, workdir, prefix, overwrite, hexfloat)
        else:
            return StoredPopResult(self, workdir, prefix, hexfloat)

    def str_bnd(self):
        return self.cmaboss_sim.str_bnd()
        
    def str_cfg(self):
        return self.cmaboss_sim.str_cfg()

    def print_bnd(self, out=sys.stdout):
        """Produce the content of the bnd file associated to the simulation."""
        print(self.str_bnd(), file=out)

    def print_cfg(self, out=sys.stdout):
        """Produce the content of the cfg file associated to the simulation."""
        print(self.str_cfg(), file=out)
__all__ = ["BNetSimulation"]



# coding: utf-8
# MaBoSS (Markov Boolean Stochastic Simulator)
# Copyright (C) 2011-2018 Institut Curie, 26 rue d'Ulm, Paris, France

# MaBoSS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# MaBoSS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# Module: maboss/comm.py
# Authors: Eric Viara <viara@sysra.com>, Vincent Noel <contact@vincent-noel.fr>
# Date: May 2018 - February 2019

from ..results.baseresult import BaseResult
from io import StringIO
import numpy as np
import pandas as pd


class Result(BaseResult):

    def __init__(self, simulation):
        BaseResult.__init__(self, simulation)
        self._status = 0
        self._errmsg = ""
        self._stat_dist = None
        self._prob_traj = None
        self._traj = None
        self._FP = None
        self._runlog = None

    def setStatus(self, status):
        self._status = status

    def setErrorMessage(self, errmsg):
        self._errmsg = errmsg

    def setStatDist(self, data_value):
        self._stat_dist = data_value

    def setProbTraj(self, data_value):
        self._prob_traj = data_value

    def setTraj(self, data_value):
        self._traj = data_value

    def setFP(self, data_value):
        self._FP = data_value

    def setRunLog(self, data_value):
        self._runlog = data_value

    def getStatus(self):
        return self._status

    def getErrorMessage(self):
        return self._errmsg

    def getStatDist(self):
        return self._stat_dist

    def getProbTraj(self):
        return self._prob_traj

    def getTraj(self):
        return self._traj

    def getFP(self):
        return self._FP

    def getRunLog(self):
        return self._runlog

    def _get_fp_fd(self):
        return StringIO(self.getFP())

    def _get_probtraj_fd(self):
        return StringIO(self.getProbTraj())

    def _get_statdist_fd(self):
        return StringIO(self.getStatDist())

    def get_probtraj_dtypes(self):

        first_line = self.getProbTraj().split("\n", 1)[0]
        cols = first_line.split("\t")
        nb_states = int((len(cols) - 5)/3)
        dtype = {
            "Time": np.float64, "TH": np.float64, "ErrorTH": np.float64, "H": np.float64, "HD=0": np.float64,
            "State": np.str, "Proba": np.float64, "ErrorProba": np.float64
        }

        for i in range(1, nb_states):
            dtype.update({"State.%d" % i: np.str, "Proba.%d" % i: np.float64, "ErrorProba.%d" % i: np.float64})
        return dtype

    def get_fptable(self):
        """Return the content of fp.csv as a pandas dataframe."""
        if self._FP is not None:
            try:
                self.fptable = pd.read_csv(self._get_fp_fd(), sep="\t", skiprows=[0])

            except pd.errors.EmptyDataError:
                pass

        return self.fptable



"""
Class that contains the results of a MaBoSS simulation.
"""

import pandas as pd
import numpy as np

class ProbTrajResult(object):
    
    def __init__(self, output_nodes=None):

        self.output_nodes = output_nodes    
        
        self.state_probtraj = None
        self.state_probtraj_errors = None
        self.state_probtraj_full = None
        
        self.nd_probtraj = None
        self.nd_probtraj_error = None
        
        self.entropy_probtraj = None
        self.entropy_probtraj_error = None
        
        self.last_states_probtraj = None
        self.last_nodes_probtraj = None
        
        self._raw_data = None
        self._first_state_index = None

        self._raw_states = None
        self._raw_probas = None
        self._raw_errors = None
        self._raw_entropy = None
        self._raw_last_data = None

        self.indexes = None
        self.states = None
        self.nodes = None
        self.states_indexes = None
        self.nodes_indexes = None

    def get_states_probtraj(self, prob_cutoff=None):
        """
            Returns the state probability vs time, as a pandas dataframe.

            :param float prob_cutoff: returns only the states with proba > cutoff
        """
        if self.state_probtraj is None:

            raw_states = self._get_raw_states()
            raw_probas = self._get_raw_probas()
            indexes, states = self._get_indexes()
            states_indexes = self._get_states_indexes()
            
            new_data = np.zeros((len(raw_probas), len(states)))
            for i, t_probas in enumerate(raw_probas):
                for j, proba in enumerate(t_probas):
                    new_data[i, states_indexes[raw_states[i][j]]] = proba    
            
            self.state_probtraj = pd.DataFrame(
                data=new_data,
                columns=states,
                index=indexes

            )
            self.state_probtraj.sort_index(axis=1, inplace=True)

        if prob_cutoff is not None:
            maxs = self.state_probtraj.max(axis=0)
            return self.state_probtraj[maxs[maxs>prob_cutoff].index]

        return self.state_probtraj

    def get_states_probtraj_errors(self):
        """
            Returns the state probability error vs time, as a pandas dataframe.
        """
        if self.state_probtraj_errors is None:
            
            raw_states = self._get_raw_states()
            raw_errors = self._get_raw_errors()
            indexes, states = self._get_indexes()
            states_indexes = self._get_states_indexes()

            new_data = np.zeros((len(raw_errors), len(states)))
            for i, t_errors in enumerate(raw_errors):
                for j, error in enumerate(t_errors):
                    new_data[i, states_indexes[raw_states[i][j]]] = error    
            
            self.state_probtraj_errors = pd.DataFrame(
                data=new_data,
                columns=states,
                index=indexes

            )
            self.state_probtraj_errors.sort_index(axis=1, inplace=True)

        return self.state_probtraj_errors

    def get_nodes_probtraj(self, nodes=None, prob_cutoff=None):
        """
            Returns the node probability vs time, as a pandas dataframe.

            :param float prob_cutoff: returns only the nodes with proba > cutoff
        """
        if self.nd_probtraj is None:

            raw_states = self._get_raw_states()
            raw_probas = self._get_raw_probas()
            indexes, states = self._get_indexes()
            nodes = self.output_nodes if self.output_nodes is not None else self._get_nodes()
            nodes_indexes = self._get_nodes_indexes()

            new_probs = np.zeros((len(indexes), len(nodes)))
            for i, t_probas in enumerate(raw_probas):
                for j, proba in enumerate(t_probas):
                    if raw_states[i][j] != "<nil>":
                        for node in raw_states[i][j].split(" -- "):
                            new_probs[i, nodes_indexes[node]] += proba

            self.nd_probtraj = pd.DataFrame(new_probs, columns=nodes, index=indexes)
            self.nd_probtraj.sort_index(axis=1, inplace=True)

        if prob_cutoff is not None:
            maxs = self.nd_probtraj.max(axis=0)
            return self.nd_probtraj[maxs[maxs>prob_cutoff].index]

        if nodes is not None:
            return self.nd_probtraj[nodes]

        return self.nd_probtraj

    def get_nodes_probtraj_error(self):
        """
            Returns the node probability error vs time, as a pandas dataframe.
        """
        if self.nd_probtraj_error is None:

            raw_states = self._get_raw_states()
            raw_errors = self._get_raw_errors()
            indexes, states = self._get_indexes()
            nodes = self.output_nodes if self.output_nodes is not None else self._get_nodes()
            nodes_indexes = self._get_nodes_indexes()
            
            new_errors = np.zeros((len(indexes), len(nodes)))
            for i, t_raw_errors in enumerate(raw_errors):
                for j, error in enumerate(t_raw_errors):
                    if raw_states[i][j] != "<nil>":
                        for node in raw_states[i][j].split(" -- "):
                            new_errors[i, nodes_indexes[node]] += error

            self.nd_probtraj_error = pd.DataFrame(new_errors, columns=nodes, index=indexes)
            self.nd_probtraj_error.sort_index(axis=1, inplace=True)

        return self.nd_probtraj_error

    def get_states_probtraj_full(self, prob_cutoff=None):
        if self.state_probtraj_full is None:

            raw_states = self._get_raw_states()
            raw_probas = self._get_raw_probas()
            raw_errors = self._get_raw_errors()
            raw_entropy = self._get_raw_entropy()
            indexes, states = self._get_indexes()
            states_indexes = self._get_states_indexes()

            full_cols = ["TH", "ErrorTH", "H"]
            for col in states:
                full_cols.append("Prob[%s]" % col)
                full_cols.append("ErrProb[%s]" % col)

            new_data = np.zeros((len(indexes), len(full_cols)))

            for i, t_entropy in enumerate(raw_entropy):
                new_data[i, 0:3] = t_entropy

            for i, t_probas in enumerate(raw_probas):
                for j, proba in enumerate(t_probas):
                    new_data[i, 3+(states_indexes[raw_states[i][j]]*2)] = proba

            for i, t_errors in enumerate(raw_errors):
                for j, error in enumerate(t_errors):
                    new_data[i, 4+(states_indexes[raw_states[i][j]]*2)] = error

            self.state_probtraj_full = pd.DataFrame(new_data, columns=full_cols, index=indexes)

        if prob_cutoff is not None:
            maxs = self.state_probtraj_full.max(axis=0)
            cols = ["TH", "ErrorTH", "H"]

            for state in maxs[maxs > prob_cutoff].index:
                if state.startswith("Prob["):
                    cols.append(state)
                    cols.append("Err%s" % state)

            return self.state_probtraj_full[cols]            
      
        return self.state_probtraj_full

    def get_last_states_probtraj(self, as_series=False):
        """
            Returns the asymptotic state probability, as a pandas dataframe.
        """
        if self.last_states_probtraj is None:
            
            data, first_col = self._get_raw_last_data()

            states = [s for s in data[first_col::3]]
            probs = np.array([float(v) for v in data[first_col+1::3]])
            
            if not as_series:
                self.last_states_probtraj = pd.DataFrame([probs], columns=states, index=[data[0]])
                self.last_states_probtraj.sort_index(axis=1, inplace=True)
            else:
                self.last_states_probtraj = pd.Series(probs, index=states, name=data[0])
                self.last_states_probtraj.sort_index(inplace=True)
             
        return self.last_states_probtraj

    def get_last_nodes_probtraj(self, nodes=None, as_series=False):
        """
            Returns the asymptotic node probability, as a pandas dataframe.
        """
        if self.last_nodes_probtraj is None:
            data, first_col = self._get_raw_last_data()
              
            raw_states = [s for s in data[first_col::3]]
            raw_probs = np.array([float(v) for v in data[first_col+1::3]])

            if nodes is not None:
                nodes_indexes = {node:index for index, node in enumerate(nodes)}
            else:
                nodes = set()
                for state in raw_states:
                    if state != "<nil>":
                        nodes.update([node for node in state.split(" -- ")])
                nodes = list(nodes) 
                nodes_indexes = {node:index for index, node in enumerate(nodes)}

            new_probas = np.zeros((1, len(nodes)))
            for i, proba in enumerate(raw_probs):
                if raw_states[i] != "<nil>":
                    for node in raw_states[i].split(" -- "):
                        if node in nodes:
                            new_probas[0, nodes_indexes[node]] += proba

            if not as_series:
                self.last_nodes_probtraj = pd.DataFrame(new_probas, columns=nodes, index=[data[0]])
                self.last_nodes_probtraj.sort_index(axis=1, inplace=True)
            else:
                self.last_nodes_probtraj = pd.Series(new_probas[0], index=nodes, name=data[0])
                self.last_nodes_probtraj.sort_index(inplace=True)
                
        return self.last_nodes_probtraj

    def get_entropy_trajectory(self):
        """
            Returns the entropy vs time, as a pandas dataframe.
        """
        if self.entropy_probtraj is None:

            raw_entropy = self._get_raw_entropy()
            indexes, _ = self._get_indexes()
            
            new_data = np.zeros((len(raw_entropy), 2))
            for i, entropy in enumerate(raw_entropy):
                new_data[i, 0] = entropy[0]
                new_data[i, 1] = entropy[2]
            
            self.entropy_probtraj = pd.DataFrame(
                data=new_data,
                columns=["TH", "H"],
                index=indexes
            )

        return self.entropy_probtraj

    def get_entropy_trajectory_error(self):
        """
            Returns the entropy error vs time, as a pandas dataframe.
        """
        if self.entropy_probtraj_error is None:

            raw_entropy = self._get_raw_entropy()
            indexes, _ = self._get_indexes()
            
            new_data = np.zeros((len(raw_entropy), 2))
            for i, entropy in enumerate(raw_entropy):
                new_data[i, 0] = entropy[1]
                new_data[i, 1] = entropy[2]
            
            self.entropy_probtraj_error = pd.DataFrame(
                data=new_data,
                columns=["ErrorTH", "H"],
                index=indexes
            )

        return self.entropy_probtraj_error

    def get_observed_graph(self, prob_cutoff=None):
        data = pd.read_csv(self.get_observed_graph_file(), sep="\t", index_col=0).astype(np.float64)
        for state, values in data.iterrows():
            if values.sum() > 0:
                data.loc[state, :] = values/values.sum()
            
        if prob_cutoff is not None:
            data[data < prob_cutoff] = 0
            
        return data
    
    def get_observed_durations(self, prob_cutoff=None):
        data = pd.read_csv(self.get_observed_durations_file(), sep="\t", index_col=0).astype(np.float64)
            
        if prob_cutoff is not None:
            data[data < prob_cutoff] = 0
            
        return data
        
    def _get_probtraj_fd(self):
        return open(self.get_probtraj_file(), 'r')

    def _get_raw_data(self):
    
        if self._raw_data is None:
            with self._get_probtraj_fd() as probtraj:

                raw_lines = probtraj.readlines()

                if self._first_state_index is None:                    
                    self._first_state_index = next(i for i, col in enumerate(raw_lines[0].strip("\n").split("\t")) if col == "State")

                self._raw_data = [line.strip("\n").split("\t") for line in raw_lines[1:]]
        
        return self._raw_data, self._first_state_index
        
    def _get_raw_last_data(self):
    
        if self._raw_last_data is None:
            with self._get_probtraj_fd() as probtraj:

                if self._first_state_index is None:
                    first_line = probtraj.readline()
                    self._first_state_index = next(i for i, col in enumerate(first_line.strip("\n").split("\t")) if col == "State")

                last_line = probtraj.readlines()[-1]
                self._raw_last_data = last_line.strip("\n").split("\t")
        
        return self._raw_last_data, self._first_state_index

    def _get_raw_states(self):
        if self._raw_states is None:
            data, first_state_index = self._get_raw_data()
            self._raw_states = [[s for s in t_data[first_state_index::3]] for t_data in data]
        return self._raw_states

    def _get_raw_probas(self):
        if self._raw_probas is None:
            data, first_state_index = self._get_raw_data()
            self._raw_probas = [[np.float64(p) for p in t_data[first_state_index+1::3]] for t_data in data]
        return self._raw_probas

    def _get_raw_errors(self):
        if self._raw_errors is None:
            data, first_state_index = self._get_raw_data()
            self._raw_errors = [[np.float64(p) for p in t_data[first_state_index+2::3]] for t_data in data]
        return self._raw_errors
        
    def _get_raw_entropy(self):
        if self._raw_entropy is None:
            data, _ = self._get_raw_data()
            self._raw_entropy = [[np.float64(p) for p in t_data[1:4]] for t_data in data]
        return self._raw_entropy

    def _get_indexes(self):

        if self.indexes is None:
            data, _ = self._get_raw_data()
            self.indexes = [float(t_data[0]) for t_data in data] 
            
        if self.states is None:
            self.states = set()
            for t_states in self._get_raw_states():
                self.states.update(t_states)
            self.states = list(self.states)

        return self.indexes, self.states
    
    def _get_nodes(self):
        
        if self.nodes is None:
            self.nodes = set()
            for t_states in self._get_raw_states():
                for state in t_states:
                    if state != "<nil>":
                        self.nodes.update([node for node in state.split(" -- ")])
            self.nodes = list(self.nodes)
        
        return self.nodes

    def _get_states_indexes(self):
        if self.states_indexes is None:
            _, states = self._get_indexes()
            self.states_indexes = {state:index for index, state in enumerate(states)}
        return self.states_indexes

    def _get_nodes_indexes(self):
        if self.nodes_indexes is None:
            nodes = self.output_nodes if self.output_nodes is not None else self._get_nodes()
            self.nodes_indexes = {node:index for index, node in enumerate(nodes)}
        return self.nodes_indexes

        


import platform
import os
import stat
import shutil
import sys
import zipfile
from urllib.request import urlretrieve



def get_url(os_type, arch):
    base_url = "https://github.com/sysbio-curie/MaBoSS/releases/latest/download/"
    if os_type == 'linux':
        return base_url + "MaBoSS-linux64.zip"
    elif os_type == 'darwin':
        if arch == 'arm64':
            return base_url + "MaBoSS-osx-arm64.zip"
        elif arch == 'x86_64':
            return base_url + "MaBoSS-osx64.zip"
        else:
            print(f"Unsupported architecture: {os_type} : {arch}")
            sys.exit(1)
    elif os_type.startswith('win'):
        return base_url + "MaBoSS-win64.zip"
    else:
        print(f"Unsupported OS: {os_type}")
        sys.exit(1)
        
def ensure_directory_exists(path):
    """Create directory if it does not exist."""
    if not os.path.exists(path):
        os.makedirs(path)
        
def download_and_extract(url, dest_path):
    """Download and extract a package from a URL."""
    # Ensure the destination directory exists
    ensure_directory_exists(dest_path)

    filename = url.split('/')[-1]
    file_path = os.path.join(dest_path, filename)

    # Download the file
    print(f"Downloading {filename} from {url}...")
    urlretrieve(url, file_path)

    # Extract the file
    print(f"Extracting {filename}...")
    with zipfile.ZipFile(file_path, 'r') as zip_ref:
        zip_ref.extractall(dest_path)

    for item in os.listdir(dest_path):
        st = os.stat(os.path.join(dest_path, item))
        os.chmod(os.path.join(dest_path, item), st.st_mode | stat.S_IEXEC)

    # Clean up
    os.remove(file_path)
    print(f"{filename} installed successfully.\n")
    
def is_installed(progname):
    return shutil.which(progname) is not None

def get_bin_path(os_type):
    
    if platform.system() == "Windows":
        bin_path = os.path.join(os.getenv("APPDATA"), "maboss", "bin")
    else:
        bin_path = os.path.join(os.path.expanduser("~"), ".local", "share", "maboss", "bin")

    if not os.path.exists(bin_path):
        os.makedirs(bin_path)
    
    if platform.system() == "Windows":
        os.environ["PATH"] = "%s;%s" % (bin_path, os.environ["PATH"])
    else:
        os.environ["PATH"] = "%s:%s" % (bin_path, os.environ["PATH"])
    return bin_path 

os_type = platform.system().lower()
arch = platform.machine().lower()

url = get_url(os_type, arch)
dest_path = get_bin_path(os_type)

if not is_installed("MaBoSS"):
    download_and_extract(url, dest_path)
    
print("MaBoSS is installed and ready to use.")
print(shutil.which("MaBoSS" if platform.system() != "Windows" else "MaBoSS.exe"))


"""User interface for MaBoSS."""


import subprocess
from sys import stderr
from contextlib import ExitStack
from .simulation import Simulation
from .types import Node, Network
from .figures import plot_trajectory

def export_and_run(simul, output='a'):
    """
    Print simul as in a .bnd and a .cfg file and run MaBoSS on it.

    MaBoSS needs to be accessible from your $PATH for this function to run.
    """
    with ExitStack() as stack: 
        bnd_file = stack.enter_context(open(output + '.bnd', 'w'))
        cfg_file = stack.enter_context(open(output + '.cfg', 'w'))
        simul.print_bnd(out=bnd_file)
        simul.print_cfg(out=cfg_file)

    err = subprocess.call(["MaBoSS", "-c", cfg_file.name, "-o", output,
                     bnd_file.name])
    if err:
        print("Error, MaBoSS returned non 0 value", file=stderr)
    else:
        print("MaBoSS returned 0", file=stderr)



"""Tools to draw figures from the MaBoSS results."""

import operator
import sys
import pandas as pd
import matplotlib.pylab as plt
import numpy as np
import networkx as nx
import matplotlib as mpl

def persistent_color(palette, state):

    reordered = " -- ".join(sorted(state.split(" -- ")))
    if state in palette:
        return palette[state]
    if reordered in palette:
        return palette[reordered]
    if state == "Others":
        color = "lightgray"
    else:
        count = len(palette)
        if "Others" not in palette:
            count += 1
        color = "C%d" % (count % 10)
    palette[state] = color
    palette[reordered] = color
    return color

def register_states_for_color(palette, collection):
    [persistent_color(palette, state) for state in collection]

def make_plot_trajectory(time_table, ax, palette, legend=True, error_table=None):
    register_states_for_color(palette, time_table.columns.values)
    color_list = [persistent_color(palette, idx) \
                    for idx in time_table.columns.values]
    time_table.plot(ax=ax, color=color_list, legend=legend, yerr=error_table)
    if legend:
        plt.legend(loc=(1.1, 0))


def plot_node_prob(time_table, ax, palette, legend=True, error_table=None):
    """Plot the probability of each node being up over time."""
    
    register_states_for_color(palette, time_table.columns.values)
    color_list = [persistent_color(palette, idx) \
                    for idx in time_table.columns.values]
    if len(time_table.columns) == 0:
        print("No data to plot !", file=sys.stderr)
        
    else:        
        if ax is None:
            _, ax = plt.subplots(1,1)
        
    
        time_table.plot(ax=ax, color=color_list, legend=legend, yerr=error_table)
        if not np.any(time_table):
            plt.ylim(0, 1)
        if legend:
            plt.legend(loc='upper right')
        return ax


def plot_piechart(plot_table, ax, palette, embed_labels=False, autopct=4, \
                    prob_cutoff=0.01, legend=True, nil_label=None):
    plot_line = plot_table.iloc[-1].rename("")  # Takes the last time point

    others = plot_line[plot_line <= prob_cutoff].sum()

    plot_line = plot_line[plot_line > prob_cutoff]
    plot_line.sort_values(ascending=False, inplace=True)

    if others:
        plot_line.at["Others"] = others

    register_states_for_color(palette, plot_line.index)

    plotting_labels = []
    color_list = []
    for value_index, value in enumerate(plot_line):
        state = plot_line.index.values[value_index]
        color_list.append(persistent_color(palette, state))
        if embed_labels and value >= 0.1:
            if state == "<nil>" and nil_label is not None:
                plotting_labels.append(nil_label)
            else:
                plotting_labels.append(state)
        else:
            plotting_labels.append("")

    opts = {}
    if autopct:
        cutoff = autopct if type(autopct) is not bool else 4
        opts.update(autopct=lambda p: '%1.1f%%' % p if p >= cutoff else "")
    else:
        opts.update(labeldistance=0.4)

    if others:
        plot_line = plot_line.rename({"Others": "Others (%1.2f%%)" % (others*100)})

    ax.pie(plot_line, labels=plotting_labels, radius=1.2,
           startangle=90, colors=color_list, **opts)
    ax.axis('equal')
    if legend:
        if nil_label is not None:
            ax.legend([nil_label if label == "<nil>" else label for label in plot_line.index.values], loc=(0.9, 0.2))
        else:
            ax.legend(plot_line.index.values, loc=(0.9, 0.2))


def plot_fix_point(table, ax, palette):
    """Plot a piechart representing the fixed point probability."""
    palette['no_fp'] = '#121212'
    prob_list = []
    color_list = []
    labels = []
    for i in range(len(table)):
        prob = table['Proba'][i]
        state = table['State'][i]
        prob_list.append(prob)
        color_list.append(persistent_color(palette, state))
        labels.append('FP '+str(i+1))
    prob_ns = 1 - sum(prob_list)
    if prob_ns > 0.01:
        prob_list.append(prob_ns)
        color_list.append(persistent_color(palette, "no_fp"))
        labels.append('no_fp')
    ax.pie(prob_list, labels=labels, colors=color_list)

def plot_observed_graph(table, ax, prune=True):

    if prune:
        pruned_table = table.copy()
        for i in range(len(table)):
            if table.iloc[i,:].sum() == 0 and table.iloc[:,i].sum() == 0:
                pruned_table.drop(table.columns[i], axis=1, inplace=True)
                pruned_table.drop(table.index[i], axis=0, inplace=True)
        table = pruned_table

    G = nx.from_pandas_adjacency(table,  create_using=nx.DiGraph())

    edge_colors = [edge['weight'] for _, edge in G.edges.items()]

    pos = nx.drawing.nx_agraph.graphviz_layout(G, prog="neato")
    # fig = plt.figure(figsize=(12,6), dpi=100)

    nx.draw(G, pos, with_labels=True, edgelist=[], node_size=0, ax=ax)
    cmap = mpl.colors.ListedColormap(plt.cm.Blues(np.linspace(0.2, 1, 100)))

    edges = nx.draw_networkx_edges(
        G,
        pos,
        arrowstyle="->",
        arrowsize=10,
        edgelist=G.edges().keys(),
        edge_color=edge_colors,
        edge_cmap=cmap,
        width=2,
        connectionstyle='arc3, rad = 0.1',
        ax=ax
    )

    pc = mpl.collections.PatchCollection(edges, cmap=cmap)
    pc.set_array(edge_colors)
    plt.colorbar(pc, ax=ax)




"""
Class that contains the results of a MaBoSS simulation.
"""
import numpy as np
import os
from .popresult import PopMaBoSSResult
import itertools
class StoredPopResult(PopMaBoSSResult):
    
    def __init__(self, sim, workdir=None, prefix="res", hexfloat=True):

        PopMaBoSSResult.__init__(self, sim)

        self._sim = sim

        self.raw_states_probtraj = None
        self.raw_last_states_probtraj = None
        
        self.raw_simple_probtraj = None
        self.raw_simple_last_probtraj = None

        self.state_probtraj = None
        self.state_probtraj_errors = None


        self._workdir = workdir
        self._prefix = prefix
        
        
        self._raw_data = None
        self._first_state_index = None

        self._raw_states = None
        self._raw_probas = None
        self._raw_errors = None
        self._raw_entropy = None
        self._raw_last_data = None
        self._raw_custom_last_data = None

        self.indexes = None
        self.states = None
        self.nodes = None
        self.states_indexes = None
        self.nodes_indexes = None
        self._hexfloat = hexfloat
        
        
        
        self._simple_raw_data = None
        self._simple_first_state_index = None
        self._simple_raw_states = None
        self._simple_raw_probas = None
        self._simple_raw_errors = None
        self.simple_state_probtraj = None
        self.simple_state_probtraj_errors = None
        self.simple_nd_probtraj = None
        self.simple_nd_probtraj_error = None
        self.simple_indexes = None
        self.simple_states = None
        self.simple_states_indexes = None
        self.simple_nodes = None
        self.simple_nodes_indexes = None
        self._simple_raw_popsize = None
        self._simple_raw_popsize_errors = None
        self.simple_popsize = None
        self.simple_popsize_errors = None
        
        
    def get_raw_last_states_probtraj(self):
                
        data, first_col, hexfloat = self._get_raw_last_data()

        time = data[0]
        states = [s for s in data[first_col::3]]
        if hexfloat:
            probs = np.array([float.fromhex(v) for v in data[first_col+1::3]])
        else:
            probs = np.array([float(v) for v in data[first_col+1::3]])
        
        return ([probs], [time], states)
        
    def get_raw_states_probtraj(self):
    
        raw_states = self._get_raw_states()
        raw_probas = self._get_raw_probas()
        raw_errors = self._get_raw_errors()
        indexes, states = self._get_indexes()
        states_indexes = self._get_states_indexes()
        
        new_data = np.zeros((len(raw_probas), len(states)))
        for i, t_probas in enumerate(raw_probas):
            for j, proba in enumerate(t_probas):
                new_data[i, states_indexes[raw_states[i][j]]] = proba    
        
        new_errors = np.zeros((len(raw_errors), len(states)))
        for i, t_probas in enumerate(raw_errors):
            for j, proba in enumerate(t_probas):
                new_errors[i, states_indexes[raw_states[i][j]]] = proba    
            
        return new_data, indexes, states, new_errors

    def get_raw_states_probtraj_by_index(self, index):
        
        raw_line, first_index, hexfloat = self._get_raw_data_by_index(index)
        time = None
        states = []
        for i, token in enumerate(split_tab_gen(raw_line)):
            if i == 0:
                time = token
            elif i >= self._first_state_index and (i-first_index) % 3 == 0:
                states.append(token)
                
        states_indexes = {state:i for i, state in enumerate(states)}
                
        new_data = np.zeros((len(states)))
        raw_data = split_tab_gen(raw_line)
        state = None
        value = None
        for i, token in enumerate(raw_data):
            if i >= first_index and (i-first_index) % 3 == 0:
                state = token
            elif i >= first_index and (i-first_index) % 3 == 1:
                value = token
                if hexfloat:
                    new_data[states_indexes[state]] = float.fromhex(value)
                else:
                    new_data[states_indexes[state]] = float(value)
                    
        return [new_data], [time], states
    ########### Simple Last Probtraj

    def get_raw_simple_last_probtraj(self):
        if self.raw_simple_last_probtraj is None:
            self.raw_simple_last_probtraj = self.cmaboss_result.get_simple_last_probtraj()
        return self.raw_simple_last_probtraj

    ########### Simple Probtraj
    
    def get_raw_simple_probtraj(self):
        if self.raw_simple_probtraj is None:
            self.raw_simple_probtraj = self.cmaboss_result.get_simple_probtraj()
        return self.raw_simple_probtraj

    ########### Custom Probtraj
    
    def get_raw_custom_last_probtraj(self):
        data, first_col, hexfloat = self._get_raw_custom_last_data()

        time = data[0]
        states = [s for s in data[first_col::3]]
        if hexfloat:
            probs = np.array([float.fromhex(v) for v in data[first_col+1::3]])
        else:
            probs = np.array([float(v) for v in data[first_col+1::3]])
        
        return ([probs], [time], states)

    def _get_raw_custom_last_data(self):
        if self._raw_custom_last_data is None:
            with self._get_custom_probtraj_fd() as probtraj:

                if self._first_state_index is None:
                    first_line = probtraj.readline()
                    self._first_state_index = next(i for i, col in enumerate(first_line.strip("\n").split("\t")) if col == "State")

                last_line = probtraj.readlines()[-1]
                self._raw_custom_last_data = last_line.strip("\n").split("\t")
                self._hexfloat = self._raw_custom_last_data[self._first_state_index+1].startswith("0x")
        
        return self._raw_custom_last_data, self._first_state_index, self._hexfloat

    

    def get_probtraj_file(self):
        return os.path.join(self._workdir, "%s_pop_probtraj.csv" % self._prefix)

    def get_simple_probtraj_file(self):
        return os.path.join(self._workdir, "%s_simple_pop_probtraj.csv" % self._prefix)
    
    def get_custom_probtraj_file(self):
        return os.path.join(self._workdir, "%s_custom_pop_probtraj.csv" % self._prefix)
    
    def _get_probtraj_fd(self):
        return open(self.get_probtraj_file(), 'r')

    def _get_custom_probtraj_fd(self):
        return open(self.get_custom_probtraj_file(), 'r')

    def _get_raw_data(self):
    
        if self._raw_data is None:
            with self._get_probtraj_fd() as probtraj:

                raw_lines = probtraj.readlines()

                if self._first_state_index is None:                    
                    self._first_state_index = next(i for i, col in enumerate(raw_lines[0].strip("\n").split("\t")) if col == "State")

                self._raw_data = [line.strip("\n").split("\t") for line in raw_lines[1:]]
                self._hexfloat = self._raw_data[0][self._first_state_index+1].startswith("0x")

        return self._raw_data, self._first_state_index, self._hexfloat
        
    def _get_raw_last_data(self):
    
        if self._raw_last_data is None:
            with self._get_probtraj_fd() as probtraj:

                if self._first_state_index is None:
                    first_line = probtraj.readline()
                    self._first_state_index = next(i for i, col in enumerate(first_line.strip("\n").split("\t")) if col == "State")

                last_line = probtraj.readlines()[-1]
                self._raw_last_data = last_line.strip("\n").split("\t")
                self._hexfloat = self._raw_last_data[self._first_state_index+1].startswith("0x")
        
        return self._raw_last_data, self._first_state_index, self._hexfloat

    def _get_raw_data_by_index(self, index):
    
        with self._get_probtraj_fd() as probtraj:

            for i, line in enumerate(probtraj):
                if i == 0 and self._first_state_index is None:
                    self._get_first_state_index(line)
                if i == index+1:
                    raw_data = split_tab_gen(line)
                    
                    first_value = next(itertools.islice(raw_data, self._first_state_index+1, None))
                    self._hexfloat = first_value.startswith("0x")
        
                    return line, self._first_state_index, self._hexfloat


    def _get_raw_states(self):
        if self._raw_states is None:
            data, first_state_index, _ = self._get_raw_data()
            self._raw_states = [[s for s in t_data[first_state_index::3]] for t_data in data]
        return self._raw_states

    def _get_raw_probas(self):
        if self._raw_probas is None:
            data, first_state_index, hexfloat = self._get_raw_data()
            if hexfloat:
                self._raw_probas = [[float.fromhex(p) for p in t_data[first_state_index+1::3]] for t_data in data]
            else:
                self._raw_probas = [[np.float64(p) for p in t_data[first_state_index+1::3]] for t_data in data]
        return self._raw_probas

    def _get_raw_errors(self):
        if self._raw_errors is None:
            data, first_state_index, hexfloat = self._get_raw_data()
            if hexfloat:
                self._raw_errors = [[float.fromhex(p) for p in t_data[first_state_index+2::3]] for t_data in data]
            else:
                self._raw_errors = [[np.float64(p) for p in t_data[first_state_index+2::3]] for t_data in data]
        return self._raw_errors
        
    def _get_raw_entropy(self):
        if self._raw_entropy is None:
            data, _, hexfloat = self._get_raw_data()
            if hexfloat:
                self._raw_entropy = [[float.fromhex(p) for p in t_data[1:4]] for t_data in data]
            else:
                self._raw_entropy = [[np.float64(p) for p in t_data[1:4]] for t_data in data]
        return self._raw_entropy

    def _get_indexes(self):

        if self.indexes is None:
            data, _, hexfloat = self._get_raw_data()
            self.indexes = [float(t_data[0]) for t_data in data] 
            
        if self.states is None:
            self.states = set()
            for t_states in self._get_raw_states():
                self.states.update(t_states)
            self.states = list(self.states)

        return self.indexes, self.states
    
    def _get_first_state_index(self, line):
        if self._first_state_index is None:
            self._first_state_index = next(i for i, col in enumerate(split_tab_gen(line)) if col == "State")
        return self._first_state_index
    
    def _get_indexes_nocache(self):

        states = set()
        indexes = []
        with self._get_probtraj_fd() as probtraj:

            for line in probtraj.readlines()[1:]:
                raw_data = line.strip("\n").split("\t")
                
                indexes.append(float(raw_data[0]))
                states.update([s for s in raw_data[first_state_index::3]])
                
        return indexes, list(states)
    
    def _get_nodes(self):
        
        if self.nodes is None:
            self.nodes = set()
            for t_states in self._get_raw_states():
                for state in t_states:
                    if state != "<nil>":
                        self.nodes.update([node for node in state.split(" -- ")])
            self.nodes = list(self.nodes)
        
        return self.nodes

    def _get_states_indexes(self):
        if self.states_indexes is None:
            _, states = self._get_indexes()
            self.states_indexes = {state:index for index, state in enumerate(states)}
        return self.states_indexes

    def _get_nodes_indexes(self):
        if self.nodes_indexes is None:
            nodes = self.output_nodes if self.output_nodes is not None else self._get_nodes()
            self.nodes_indexes = {node:index for index, node in enumerate(nodes)}
        return self.nodes_indexes

        
        
        
        
        
        
        
        
    def _get_simple_probtraj_fd(self):
        return open(self.get_simple_probtraj_file(), 'r')

    def _get_simple_raw_data(self):
    
        if self._simple_raw_data is None:
            with self._get_simple_probtraj_fd() as probtraj:

                raw_lines = probtraj.readlines()

                if self._simple_first_state_index is None:                    
                    self._simple_first_state_index = next(i for i, col in enumerate(raw_lines[0].strip("\n").split("\t")) if col == "State")
                
                self._simple_raw_data = [line.strip("\n").split("\t") for line in raw_lines[1:]]
                self._hexfloat = self._simple_raw_data[0][self._simple_first_state_index+1].startswith("0x")

        return self._simple_raw_data, self._simple_first_state_index, self._hexfloat
        
    def _get_simple_raw_popsize(self):
        if self._simple_raw_popsize is None:
            data, _, hexfloat = self._get_simple_raw_data()
            if hexfloat:
                self._simple_raw_popsize = [float.fromhex(t_data[5]) for t_data in data]
            else:
                self._simple_raw_popsize = [np.float64(t_data[5]) for t_data in data]
    
        return self._simple_raw_popsize
    
    def _get_simple_raw_popsize_errors(self):
        if self._simple_raw_popsize_errors is None:
            data, _, hexfloat = self._get_simple_raw_data()
            if hexfloat:
                self._simple_raw_popsize_errors = [float.fromhex(t_data[6]) for t_data in data]
            else:
                self._simple_raw_popsize_errors = [np.float64(t_data[6]) for t_data in data]
    
        return self._simple_raw_popsize_errors
        
    def _get_simple_raw_states(self):
        if self._simple_raw_states is None:
            data, first_state_index, _ = self._get_simple_raw_data()
            self._simple_raw_states = [[s for s in t_data[first_state_index::3]] for t_data in data]
        return self._simple_raw_states

    def _get_simple_raw_probas(self):
        if self._simple_raw_probas is None:
            data, first_state_index, hexfloat = self._get_simple_raw_data()
            if hexfloat:
                self._simple_raw_probas = [[float.fromhex(p) for p in t_data[first_state_index+1::3]] for t_data in data]
            else:
                self._simple_raw_probas = [[np.float64(p) for p in t_data[first_state_index+1::3]] for t_data in data]
        return self._simple_raw_probas

    def _get_simple_raw_errors(self):
        if self._simple_raw_errors is None:
            data, first_state_index, hexfloat = self._get_simple_raw_data()
            if hexfloat:
                self._simple_raw_errors = [[float.fromhex(p) for p in t_data[first_state_index+2::3]] for t_data in data]
            else:
                self._simple_raw_errors = [[np.float64(p) for p in t_data[first_state_index+2::3]] for t_data in data]
        return self._simple_raw_errors

    def _get_simple_indexes(self):

        if self.simple_indexes is None:
            data, _, _ = self._get_simple_raw_data()
            self.simple_indexes = [float(t_data[0]) for t_data in data] 
            
        if self.simple_states is None:
            self.simple_states = set()
            for t_states in self._get_simple_raw_states():
                self.simple_states.update(t_states)
            self.simple_states = list(self.simple_states)

        return self.simple_indexes, self.simple_states
    
    def _get_simple_states_indexes(self):
        if self.simple_states_indexes is None:
            _, states = self._get_simple_indexes()
            self.simple_states_indexes = {state:index for index, state in enumerate(states)}
        return self.simple_states_indexes
    
    def _get_simple_nodes(self):
        
        if self.simple_nodes is None:
            self.simple_nodes = set()
            for t_states in self._get_simple_raw_states():
                for state in t_states:
                    if state != "<nil>":
                        self.simple_nodes.update([node for node in state.split(" -- ")])
            self.simple_nodes = list(self.simple_nodes)
        
        return self.simple_nodes

    def _get_simple_nodes_indexes(self):
        if self.simple_nodes_indexes is None:
            nodes = self.output_nodes if self.output_nodes is not None else self._get_simple_nodes()
            self.simple_nodes_indexes = {node:index for index, node in enumerate(nodes)}
        return self.simple_nodes_indexes
        
    def get_raw_simple_probtraj(self, prob_cutoff=None):
        """
            Returns the simplified population state probability vs time, as a pandas dataframe.

            :param float prob_cutoff: returns only the states with proba > cutoff
        """

        raw_states = self._get_simple_raw_states()
        raw_probas = self._get_simple_raw_probas()
        raw_errors = self._get_simple_raw_errors()
        raw_popsizes = self._get_simple_raw_popsize()
        raw_popsizes_errors = self._get_simple_raw_popsize_errors()
        
        indexes, states = self._get_simple_indexes()
        states_indexes = self._get_simple_states_indexes()
        new_data = np.zeros((len(raw_probas), len(states)+1))
        for i, t_probas in enumerate(raw_probas):
            new_data[i, 0] = raw_popsizes[i]
            for j, proba in enumerate(t_probas):
                new_data[i, states_indexes[raw_states[i][j]]+1] = proba    
        
        new_error = np.zeros((len(raw_errors), len(states)+1))
        for i, t_probas in enumerate(raw_errors):
            new_error[i, 0] = raw_popsizes_errors[i]
            for j, proba in enumerate(t_probas):
                new_error[i, states_indexes[raw_states[i][j]]+1] = proba    
                
        return new_data, indexes, ["Population"] + states, new_error
    
    # def get_simple_states_probtraj_errors(self, prob_cutoff=None):
    #     """
    #         Returns the simplified population state probability vs time, as a pandas dataframe.

    #         :param float prob_cutoff: returns only the states with proba > cutoff
    #     """
    #     if self.simple_state_probtraj_errors is None:

    #         raw_states = self._get_simple_raw_states()
    #         raw_errors = self._get_simple_raw_errors()
            
    #         indexes, states = self._get_simple_indexes()
    #         states_indexes = self._get_simple_states_indexes()
    #         new_data = np.zeros((len(raw_errors), len(states)))
    #         for i, t_probas in enumerate(raw_errors):
    #             for j, proba in enumerate(t_probas):
    #                 new_data[i, states_indexes[raw_states[i][j]]] = proba    
            
    #         self.simple_state_probtraj_errors = pd.DataFrame(
    #             data=new_data,
    #             columns=states,
    #             index=indexes

    #         )
    #         self.simple_state_probtraj_errors.sort_index(axis=1, inplace=True)

    #     if prob_cutoff is not None:
    #         maxs = self.simple_state_probtraj_errors.max(axis=0)
    #         return self.simple_state_probtraj_errors[maxs[maxs>prob_cutoff].index]

    #     return self.simple_state_probtraj_errors


    # def get_simple_nodes_probtraj(self, nodes=None, prob_cutoff=None):
    #     """
    #         Returns the node probability vs time, as a pandas dataframe.

    #         :param float prob_cutoff: returns only the nodes with proba > cutoff
    #     """
    #     if self.simple_nd_probtraj is None:

    #         raw_states = self._get_simple_raw_states()
    #         raw_probas = self._get_simple_raw_probas()
    #         indexes, states = self._get_simple_indexes()
    #         nodes = self.output_nodes if self.output_nodes is not None else self._get_simple_nodes()
    #         nodes_indexes = self._get_simple_nodes_indexes()

    #         new_probs = np.zeros((len(indexes), len(nodes)))
    #         for i, t_probas in enumerate(raw_probas):
    #             for j, proba in enumerate(t_probas):
    #                 if raw_states[i][j] != "<nil>":
    #                     for node in raw_states[i][j].split(" -- "):
    #                         new_probs[i, nodes_indexes[node]] += proba

    #         self.simple_nd_probtraj = pd.DataFrame(new_probs, columns=nodes, index=indexes)
    #         self.simple_nd_probtraj.sort_index(axis=1, inplace=True)

    #     if prob_cutoff is not None:
    #         maxs = self.simple_nd_probtraj.max(axis=0)
    #         return self.simple_nd_probtraj[maxs[maxs>prob_cutoff].index]

    #     if nodes is not None:
    #         return self.simple_nd_probtraj[nodes]

    #     return self.simple_nd_probtraj

    # def get_simple_nodes_probtraj_error(self):
    #     """
    #         Returns the node probability error vs time, as a pandas dataframe.
    #     """
    #     if self.simple_nd_probtraj_error is None:

    #         raw_states = self._get_simple_raw_states()
    #         raw_errors = self._get_simple_raw_errors()
    #         indexes, states = self._get_simple_indexes()
    #         nodes = self.output_nodes if self.output_nodes is not None else self._get_simple_nodes()
    #         nodes_indexes = self._get_simple_nodes_indexes()
            
    #         new_errors = np.zeros((len(indexes), len(nodes)))
    #         for i, t_raw_errors in enumerate(raw_errors):
    #             for j, error in enumerate(t_raw_errors):
    #                 if raw_states[i][j] != "<nil>":
    #                     for node in raw_states[i][j].split(" -- "):
    #                         new_errors[i, nodes_indexes[node]] += error

    #         self.simple_nd_probtraj_error = pd.DataFrame(new_errors, columns=nodes, index=indexes)
    #         self.simple_nd_probtraj_error.sort_index(axis=1, inplace=True)

    #     return self.simple_nd_probtraj_error


    # def get_simple_popsize(self):
    #     if self.simple_popsize is None:
    #         indexes, _ = self._get_simple_indexes()
    #         popsizes = self._get_simple_raw_popsize()
            
    #         self.simple_popsize = pd.Series(popsizes, index=indexes, name="Popsize")
    #     return self.simple_popsize
    
    # def get_simple_popsize_errors(self):
    #     if self.simple_popsize_errors is None:
    #         indexes, _ = self._get_simple_indexes()
    #         popsizes_errors = self._get_simple_raw_popsize_errors()
            
    #         self.simple_popsize_errors = pd.Series(popsizes_errors, index=indexes, name="Popsize")
    #     return self.simple_popsize_errors
           
    # def get_simple_states_popsize(self):
    #     return self.get_simple_states_probtraj().multiply(self.get_simple_popsize(), axis=0)
       
       
    # def get_simple_nodes_popsize(self, nodes=None):
    #     return self.get_simple_nodes_probtraj(nodes).multiply(self.get_simple_popsize(), axis=0)
       
def split_tab_gen(string):
    start = 0
    for i, c in enumerate(string):
        if c == "\t":
            yield string[start:i]
            start = i+1
    yield string[start:]



from ctypes.util import find_library
from colomoto.setup_helper import setup
setup({"pkg": "colomoto/maboss",
    "check_progs": ["MaBoSS"]},
    {"pkg": "colomoto/libsbml-plus-packages",
        "check_install": lambda: find_library("sbml")})



"""
Class that contains the results of a PopMaBoSS simulation.
"""

import pandas, numpy, os

class PopMaBoSSResult:
            
    def __init__(self, sim):
        self._sim = sim
        
    def get_last_states_probtraj(self):
        
        raw_res = self.get_raw_last_states_probtraj()
        df = pandas.Series(
            raw_res[0][0], 
            index=raw_res[2], 
            name=raw_res[1][0]
        )
        df.sort_index(inplace=True)

        return df

    def get_states_probtraj(self, prob_cutoff=None):

        raw_res = self.get_raw_states_probtraj()
        df = pandas.DataFrame(*(raw_res[0:3]))
        # df.sort_index(axis=1, inplace=True)

        if prob_cutoff is not None:
            maxs = df.max(axis=0)
            return df[maxs[maxs>prob_cutoff].index]

        return df

    def get_states_probtraj_by_index(self, index):
        
        raw_res = self.get_raw_states_probtraj_by_index(index)
        df = pandas.Series(
            raw_res[0][0], 
            index=raw_res[2], 
            name=raw_res[1][0]
        )
        df.sort_index(inplace=True)

        return df

    ########### Simple Last Probtraj

    def get_last_simple_states_probtraj(self):
        raw_res = self.get_raw_simple_probtraj()
        df = pandas.Series(raw_res[0][1:], index=raw_res[1], name=raw_res[2][1:])
        df.sort_index(inplace=True)
        return df
    
    def get_last_simple_states_probtraj_errors(self):
        raw_res = self.get_raw_simple_probtraj()
        df = pandas.Series(raw_res[3][1:], index=raw_res[1], name=raw_res[2][1:])
        df.sort_index(inplace=True)
        return df

    ########### Simple Probtraj
    
    def get_simple_states_probtraj(self):
        raw_res = self.get_raw_simple_probtraj()
        df = pandas.DataFrame(raw_res[0][:,1:], index=raw_res[1], columns=raw_res[2][1:])
        df.sort_index(axis=1, inplace=True)
        return df

    def get_simple_states_probtraj_errors(self):
        raw_res = self.get_raw_simple_probtraj()
        df = pandas.DataFrame(raw_res[3][:,1:], index=raw_res[1], columns=raw_res[2][1:])
        df.sort_index(axis=1, inplace=True)
        return df

    def get_custom_states_probtraj(self):
        raw_res = self.get_raw_custom_probtraj()
        df = pandas.DataFrame(raw_res[0][:,], index=raw_res[1], columns=[int(x) for x in raw_res[2]])
        df.sort_index(axis=1, inplace=True)
        df.sort_index(axis=0, inplace=True)
        return df
        
    def get_custom_states_probtraj_errors(self):
        raw_res = self.get_raw_custom_probtraj()
        df = pandas.DataFrame(raw_res[3][:,], index=raw_res[1], columns=[int(x) for x in raw_res[2]])
        df.sort_index(axis=1, inplace=True)
        df.sort_index(axis=0, inplace=True)
        
        return df
        
    def get_last_custom_states_probtraj(self, prob_cutoff=None, rescale=True):
        raw_res = self.get_raw_custom_last_probtraj()
        df = pandas.Series(raw_res[0][0], index=[int(x) for x in raw_res[2]], name=raw_res[1][0])
        
        if rescale:
            df /= df.sum()
        
        if prob_cutoff is not None:
            df = df[df > prob_cutoff]
        
        df.sort_index(inplace=True)
        
        return df
    
    
    ########### Popsize
    
    def get_simple_popsize(self):
        raw_res = self.get_raw_simple_probtraj()
        df = pandas.Series(raw_res[0][:,0], index=raw_res[1], name=raw_res[2][0]) 
        return df
    
    def get_simple_popsize_errors(self):
        raw_res = self.get_raw_simple_probtraj()
        df = pandas.Series(raw_res[3][:,0], index=raw_res[1], name=raw_res[2][0]) 
        return df
    
    def get_simple_states_popsize(self):
        return self.get_simple_states_probtraj().multiply(self.get_simple_popsize(), axis=0)
    
    # def get_simple_states_popsize_errors(self):
    #     return self.get_simple_states_probtraj_errors().multiply(self.get_simple_popsize_errors(), axis=0)
    
    ########### Simple nodes probtraj
    
    def get_simple_nodes_probtraj(self):
        
        raw_probas, indexes, states, _ = self.get_raw_simple_probtraj()
        nodes = self._sim.get_nodes()
        nodes_indexes = {node:index for index, node in enumerate(nodes)}
        
        nodes_probas = numpy.zeros((len(indexes), len(nodes)))
        for i in range(len(indexes)):
            for j, state in enumerate(states[1:]):
                if state != "<nil>":
                    for node in state.split(" -- "):
                        nodes_probas[i, nodes_indexes[node]] += raw_probas[i][j+1]
    
        df = pandas.DataFrame(nodes_probas, index=indexes, columns=nodes)
        return df
    
    def get_simple_nodes_errors(self):
        
        _, indexes, states, raw_errors = self.get_raw_simple_probtraj()
        nodes = self._sim.get_nodes()
        nodes_indexes = {node:index for index, node in enumerate(nodes)}
        
        nodes_errors = numpy.zeros((len(indexes), len(nodes)))
        for i in range(len(indexes)):
            for j, state in enumerate(states[1:]):
                if state != "<nil>":
                    for node in state.split(" -- "):
                        nodes_errors[i, nodes_indexes[node]] += raw_errors[i][j+1]
    
        df = pandas.DataFrame(nodes_errors, index=indexes, columns=nodes)
        return df
    
    
    ########### Nodes Probtraj
 
    def get_nodes_probtraj(self, nodes=[], prob_cutoff=None):
        
        # if len(nodes) == 0:
        #     _nodes = self.sim.get_nodes()
            
        raw_probas, indexes, states, _ = self.get_raw_states_probtraj()
                
        self._popnodes = set()  #{"{%s:%s}" % (node,0) for node in _nodes}
        for t_state in states:
            for subpopstates in t_state[1:-1].split(","):
                state, pop = subpopstates[1:-1].split(":")
                if state != "<nil>":
                    for node in state.split(" -- "):
                        self._popnodes.add("{%s:%s}" % (node, pop))
                        
                        
        def sort_popnodes(popnode):
            node, pop = popnode[1:-1].split(":")
            return (node, int(pop))
        
        self._popnodes = sorted(list(self._popnodes), key=lambda key: sort_popnodes(key))
        self._popnodes_index = {popnode:index for index, popnode in enumerate(self._popnodes)}                
        
        self._popnodes_by_nodes = {node:[] for node in self._sim.get_nodes()}
        for popnode in self._popnodes:
            node, pop = popnode[1:-1].split(":")
            self._popnodes_by_nodes[node].append(popnode)
                
        new_probs = numpy.zeros((len(indexes), len(self._popnodes)))
        for i in range(len(indexes)):
            for j, state in enumerate(states):
                for subpopstates in state[1:-1].split(","):
                    state, pop = subpopstates[1:-1].split(":")
                    if state != "<nil>":
                        for node in state.split(" -- "):
                            new_probs[i, self._popnodes_index["{%s:%s}" % (node, pop)]] += raw_probas[i][j]
        
        self.nd_probtraj = pandas.DataFrame(new_probs, columns=self._popnodes, index=indexes)
        
        _list_popnodes = []
        if len(nodes) > 0:
            for node in nodes:
                _list_popnodes += self._popnodes_by_nodes[node]
        
        if prob_cutoff is not None:
            if len(_list_popnodes) > 0:
                maxs = self.nd_probtraj[_list_popnodes].max(axis=0)
            else:
                maxs = self.nd_probtraj.max(axis=0)
             
            return self.nd_probtraj[maxs[maxs>prob_cutoff].index]

        if len(_list_popnodes) > 0:
            return self.nd_probtraj[_list_popnodes]
        
        return self.nd_probtraj

    # def get_last_nodes_probtraj(self):
    #     raw_res = self.cmaboss_result.get_last_nodes_probtraj()
        
    #     df = pandas.Series(raw_res[0][0], index=raw_res[2], name=raw_res[1][0])
    #     df.sort_index(inplace=True)
    #     return df


    # def get_fptable(self):
    #     raw_res = self.cmaboss_result.get_fp_table()

    #     df = pandas.DataFrame(["#%d" % fp for fp in sorted(raw_res.keys())], columns=["FP"])

    #     df["Proba"] = [raw_res[fp][0] for fp in sorted(raw_res.keys())]
    #     df["State"] = [raw_res[fp][1] for fp in sorted(raw_res.keys())]

    #     for node in self.simul.network.keys():
    #         df[node] = [1 if node in raw_res[fp][1].split(" -- ") else 0 for fp in sorted(raw_res.keys())]

    #     return df

    def get_state_dist_by_index(self, index, fun):
        dist_state = {}
        for raw_popstates, proba in self.get_states_probtraj_by_index(index).items():
            value = fun(raw_popstates)
            if value is not None:
                if value in dist_state.keys():
                    dist_state[value] += proba
                else:
                    dist_state[value] = proba
            
        serie = pandas.Series(dist_state).sort_index()
        serie *= 1/serie.sum()
        return serie

    def get_last_state_dist(self, networkstate=None, rescale=True):
        if networkstate is None:
            return self.get_last_states_probtraj()
            
        dist_state = {}
        for raw_popstates, value in self.get_last_states_probtraj().items():
            popstates = parse_pop_state(raw_popstates)
            if networkstate in popstates.keys():
                if popstates[networkstate] in dist_state.keys():
                    dist_state[popstates[networkstate]] += value
                else:
                    dist_state[popstates[networkstate]] = value
                    
        df = pandas.Series(dist_state).sort_index()
        if rescale:
            df /= df.sum()
        return df


    def plot_last_state_dist(self, networkstate=None, **args):
        if networkstate is None:
            self.get_last_state_dist().plot.pie(autopct='%1.1f%%', legend=False)
        else:
            self.get_last_state_dist(networkstate).plot.bar(**args)
        
    # def get_node_dist(self, node):
    #     dist_node = {}
    #     for raw_popnode, value in serie.items():
    #         state, pop = raw_popnode[1:-1].split(":")
    #         dist_node.update({int(pop): value})
    #     return pandas.Series(dist_node).sort_index()
    def get_nb_dists(self):

        raw_probas, indexes, states, _ = self.get_raw_states_probtraj()
        
        sizes = set()
        for state in states:
            popstates = parse_pop_state(state)
            nb = sum(popstates.values())
            sizes.add(nb)
            
        sizes = sorted(list(sizes))
        sizes_indexes = {size:index for index, size in enumerate(sizes)}
        
        raw_nb_dists = numpy.zeros((len(indexes), len(sizes)))
        for i in range(len(indexes)):
            for j, state in enumerate(states):
                popstates = parse_pop_state(state)
                nb = sum(popstates.values())
                raw_nb_dists[i, sizes_indexes[nb]] += raw_probas[i][j]
                
        return pandas.DataFrame(raw_nb_dists, columns=sizes, index=indexes)
    
    def get_last_nb_dist(self):
        dist_state = {}
        for raw_popstates, value in self.get_last_states_probtraj().items():
            if value > 0:
                popstates = parse_pop_state(raw_popstates)
                nb = sum(popstates.values())
                if nb in dist_state.keys():
                    dist_state[nb] += value
                else:
                    dist_state[nb] = value
        return pandas.Series(dist_state).sort_index()

    def plot_last_nb_dist(self, **xargs):
        self.get_last_nb_dist().plot.bar(**xargs)
    
    def get_last_log_nb_dist(self, num=20, vmin=None, vmax=None, lspace=None):
        nbdist = self.get_last_nb_dist()
        ldist = numpy.log(numpy.array(nbdist.index))
        values = {}
        if lspace is None:
            if vmin is None:
                lmin = numpy.floor(ldist[0])
            else:
                lmin = numpy.log(vmin)
                
            if vmax is None:
                lmax = numpy.ceil(ldist[-1])
            else:
                lmax = numpy.log(vmax)
                
            lspace = numpy.linspace(lmin, lmax, num)
        for i, lognb in enumerate(lspace):
            if i < (len(lspace)-1):
                where = numpy.where((ldist >= lognb) & (ldist < lspace[i+1]))
                proba = nbdist.iloc[where].sum()
                values[numpy.exp(lognb)] = proba
        return pandas.Series(values)

    def plot_last_log_nb_dist(self, num=20, vmin=None, vmax=None, lspace=None, **xargs):
        self.get_last_log_nb_dist(num, vmin, vmax, lspace).plot.bar(**xargs)

    ########### Activity ratios

    def get_activity_ratio_expected(self, node):
        
        raw_probas, indexes, states, _ = self.get_raw_states_probtraj()
        activity_ratio = []
        for i in range(len(indexes)):
            activity = 0
            for j, state in enumerate(states):
                popstates = parse_pop_state(state)
                nb = sum(popstates.values())
                
                for state, pop in popstates.items():
                    if state != "<nil>" and node in state.split(" -- "):                 
                        activity += raw_probas[i][j]*pop/nb
                        
            activity_ratio.append(activity)
        return pandas.Series(activity_ratio, index=indexes, name=node)
        
    def get_activity_ratio_stdev(self, node):
        
        raw_probas, indexes, states, _ = self.get_raw_states_probtraj()
        activity_ratio = self.get_activity_ratio_expected(node)
        
        stdevs = []
        for i in range(len(indexes)):
            stdev = 0
            for j, state in enumerate(states):
                    popstates = parse_pop_state(state)
                    nb = sum(popstates.values())
                    t_proba = 0
                    t_stdev = 0
                    for state, pop in popstates.items():
                        if state != "<nil>" and node in state.split(" -- "):                 
                            t_proba += raw_probas[i][j]
                            t_stdev += pop/nb
                    stdev += ((t_stdev-activity_ratio.iloc[i])**2)*t_proba            
            stdevs.append((stdev)**0.5)
        return pandas.Series(stdevs, index=indexes, name=node)
    
    def get_activity_ratio_nodes(self, node0, node1, time=None, nbins=20):
        raw_probas, indexes, states, _ = self.get_raw_states_probtraj()
        
        if time is None:
            raw_probas = raw_probas[-1]
            indexes = indexes[-1]
        else:
            raw_probas = raw_probas[indexes.index(time)]
            indexes = time
            
        space0 = numpy.linspace(0, 1, nbins+1)
        space1 = numpy.linspace(0, 1, nbins+1)
        probas = numpy.zeros((len(space0), len(space1)))
        for j, state in enumerate(states):
            popstates = parse_pop_state(state)
            nb = sum(popstates.values())
            pop0 = 0
            pop1 = 0
            for state, pop in popstates.items():
                if state != "<nil>":
                    popnb = pop/nb
                    if node0 in state.split(" -- "):                 
                        pop0 += popnb
                    if node1 in state.split(" -- "):
                        pop1 += popnb
            pos0 =  numpy.floor(pop0*nbins).astype(int)
            pos1 =  numpy.floor(pop1*nbins).astype(int)
            probas[pos0, pos1] += raw_probas[j]

        return pandas.DataFrame(probas, index=["%.2g" % val for val in space1], columns=["%.2g" % val for val in space0])
    
    def get_activity_ratio_states(self, state0, state1, time=None, nbins=20):
        raw_probas, indexes, states, _ = self.get_raw_states_probtraj()
        
        if time is None:
            raw_probas = raw_probas[-1]
            indexes = indexes[-1]
        else:
            raw_probas = raw_probas[indexes.index(time)]
            indexes = time
            
        space0 = numpy.linspace(0, 1, nbins+1)
        space1 = numpy.linspace(0, 1, nbins+1)
        probas = numpy.zeros((len(space0), len(space1)))
        for j, state in enumerate(states):
            popstates = parse_pop_state(state)
            nb = sum(popstates.values())
            pop0 = 0
            pop1 = 0
            for state, pop in popstates.items():
                if state == state0:
                    pop0 += pop/nb
                if state == state1:
                    pop1 += pop/nb
            pos0 =  numpy.floor(pop0*nbins).astype(int)
            pos1 =  numpy.floor(pop1*nbins).astype(int)
            probas[pos0, pos1] += raw_probas[j]

        return pandas.DataFrame(probas, index=["%.2g" % val for val in space1], columns=["%.2g" % val for val in space0])
    
def parse_pop_state(cols):    
    pops = {}
    if len(cols) > 2:
        for pop in cols[1:-1].split(","):
            name, value = pop[1:-1].split(":")
            pops[name] = int(value)
    return pops

__all__ = ["PopMaBoSSResult"]



"""
Class that contains the results of a MaBoSS simulation.
"""

from __future__ import print_function

from .results.baseresult import BaseResult

from sys import stderr, stdout, version_info
if version_info[0] < 3:
    from contextlib2 import ExitStack
else:
    from contextlib import ExitStack
import shutil
import tempfile
import os
import subprocess


class Result(BaseResult):

    def __init__(self, simul, command=None, workdir=None, overwrite=False, prefix="res"):

        self.workdir = workdir
        self.prefix = prefix
        if workdir is None:
            self._path = tempfile.mkdtemp()
        else:
            self._path = workdir
            if os.path.exists(self._path) and overwrite:
                shutil.rmtree(self._path)
                os.mkdir(self._path)
            
            elif not os.path.exists(self._path):
                os.mkdir(self._path)

        self.output_nodes = [name for name, node in simul.network.items() if not node.is_internal]
        BaseResult.__init__(self, self._path, simul, command, output_nodes=self.output_nodes)

        if workdir is None or len(os.listdir(workdir)) == 0 or overwrite:

            cfg_fd, self._cfg = tempfile.mkstemp(dir=self._path, suffix='.cfg')
            os.close(cfg_fd)
            
            bnd_fd, self._bnd = tempfile.mkstemp(dir=self._path, suffix='.bnd')
            os.close(bnd_fd)
            
            with ExitStack() as stack:
                bnd_file = stack.enter_context(open(self._bnd, 'w'))
                cfg_file = stack.enter_context(open(self._cfg, 'w'))
                simul.print_bnd(out=bnd_file)
                simul.print_cfg(out=cfg_file)

            maboss_cmd = simul.get_maboss_cmd()
            if command:
                maboss_cmd = command

            self._err = subprocess.call([maboss_cmd, "-c", self._cfg, "-o",
                                        os.path.join(self._path, self.prefix), self._bnd])
            if self._err:
                print("Error, MaBoSS returned non 0 value", file=stderr)

    def get_fp_file(self):
        return "{}/{}_fp.csv".format(self._path, self.prefix)

    def get_probtraj_file(self):
        return "{}/{}_probtraj.csv".format(self._path, self.prefix)
    
    def get_statdist_file(self):
        return "{}/{}_statdist.csv".format(self._path, self.prefix)
    
    def get_observed_graph_file(self):
        return "{}/{}_observed_graph.csv".format(self._path, self.prefix)
    
    def get_observed_durations_file(self):
        return "{}/{}_observed_durations.csv".format(self._path, self.prefix)
    
    def save(self, prefix, replace=False):
        """
        Write the cfg, bnd and all results in working dir.

        prefix is a string that will determine the name of the created files.
        If there is a conflict with existing files, the existing files will be
        replaced or not, depending on the value of the replace argument.
        """
        if not _check_prefix(prefix):
            return

        # Create the results directory
        try:
            os.makedirs(prefix)
        except OSError:
            if not replace:
                print('Error directory already exists: %s' % prefix,
                      file=stderr)
                return
            elif prefix.startswith('rpl_'):
                shutil.rmtree(prefix)
                os.makedirs(prefix)
            else:
                print('Error only directries begining with "rpl_" can be'
                      'replaced', file=stderr)
                return

        # Moves all the files into it
        shutil.copy(self._bnd, prefix+'/%s.bnd' % os.path.basename(prefix))
        shutil.copy(self._cfg, prefix+'/%s.cfg' % os.path.basename(prefix))

        maboss_files = filter(lambda x: x.startswith(self.prefix),
                              os.listdir(self._path))
        for f in maboss_files:
            shutil.copy(self._path + '/' + f, prefix)

    def __del__(self):
        if self.workdir is None and os.path.exists(self._path):
            shutil.rmtree(self._path)

def _check_prefix(prefix):
    if type(prefix) is not str:
        print('Error save method expected string')
        return False
    return True

__all__ = ["Result"]



"""
Class that contains the cMaBoSS simulation.
"""
from .cmabossresult2 import CMaBoSSResult2
from sys import stdout

class CMaBoSSSimulation(object):

    def __init__(self, bnd=None, cfgs=None, cmaboss=None, cmaboss_sim=None):
        
        if cmaboss is not None and cmaboss_sim is not None:
            self.cmaboss = cmaboss
            self.cmaboss_sim = cmaboss_sim
            self.network = self.cmaboss_sim.network
        else:
            self.bnd = bnd
            self.cfgs = cfgs

            self.nb_nodes = self.count_nodes()
            self.cmaboss = self.get_cmaboss()

            self.cmaboss_net = self.cmaboss.MaBoSSNet(self.bnd)
            self.cmaboss_cfg = self.cmaboss.MaBoSSCfg(self.cmaboss_net, self.cfgs)
            
            self.cmaboss_sim = self.cmaboss.MaBoSSSim(net=self.cmaboss_net, cfg=self.cmaboss_cfg)
            self.network = self.cmaboss_net    
            
        self.param = self.cmaboss_sim.param
        

    def count_nodes(self):
        
        res = 0
        with open(self.bnd, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if line.lower().startswith("node"):
                    res += 1
        return res


    def get_cmaboss(self):

        assert self.nb_nodes <= 1024, "Models with more than 1024 nodes are not compatible with this version of MaBoSS"

        if self.nb_nodes <= 64:
            return __import__("cmaboss")
        elif self.nb_nodes <= 128:
            return __import__("cmaboss_128n")
        elif self.nb_nodes <= 256:
            return __import__("cmaboss_256n")
        elif self.nb_nodes <= 512:
            return __import__("cmaboss_512n")
        else:
            return __import__("cmaboss_1024n")

    def get_logical_rules(self):

        out = self.cmaboss_sim.get_logical_rules()    
        rules = {}
        for line in out.split("\n"):
            if ":" in line:
                node, rule = line.split(" : ", 1)
                rules.update({node.strip(): rule.strip()})

        return rules
        
    def update_parameters(self, **kwargs):
        self.cmaboss_sim.update_parameters(**kwargs)

    def run(self, workdir=None, only_final_state=False):
        return CMaBoSSResult2(self, workdir, only_final_state)

    def check(self):
        return self.cmaboss.MaBoSSSim(net=self.cmaboss_net, cfg=self.cmaboss_cfg)

    def str_bnd(self):
        return self.cmaboss_sim.str_bnd()
        
    def str_cfg(self):
        return self.cmaboss_sim.str_cfg()

    def copy(self):
        return CMaBoSSSimulation(cmaboss=self.cmaboss, cmaboss_sim=self.cmaboss_sim.copy())
        
    def print_bnd(self, out=stdout):
        """Produce the content of the bnd file associated to the simulation."""
        print(self.str_bnd(), file=out)

    def print_cfg(self, out=stdout):
        """Produce the content of the cfg file associated to the simulation."""
        print(self.str_cfg(), file=out)

__all__ = ["CMaBoSSSimulation"]



from sys import version_info

from setuptools import setup, find_packages

optional_contextlib = []
if version_info[0] < 3:
    optional_contextlib.append("contextlib2")

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "description.rst").read_text()

setup(name='maboss',
    version="0.8.10",
    packages=find_packages(exclude=["test"]),
    py_modules = ["maboss_setup", "maboss_setup_experimental"],
    author="Vincent Noël, Loic Paulevé, Aurelien Naldi and Nicolas Levy",
    author_email="vincent.noel@curie.fr",
    description="A python and jupyter API for the MaBoSS software",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url="https://maboss.curie.fr",
    install_requires = [
        "colomoto_jupyter >=0.4.10",
        "pyparsing",
        "ipywidgets",
        "matplotlib",
        "pandas",
        "scikit-learn",
        "networkx",
        "cmaboss>=1.0.0b26; platform_system!='Windows'"
    ] + optional_contextlib,
    scripts=['scripts/MBSS_FormatTable.py', 'scripts/UpPMaBoSS.py']
)



"""
Class that contains the results of a PopMaBoSS simulation.
"""

import os

from .popresult import PopMaBoSSResult

class cMaBoSSPopMaBoSSResult(PopMaBoSSResult):

    def __init__(self, sim, workdir=None, prefix="res", overwrite=False, hexfloat=True):

        PopMaBoSSResult.__init__(self, sim)
        
        self._sim = sim        
        self.cmaboss_result = sim.cmaboss_sim.run()

        self.raw_states_probtraj = None
        self.raw_last_states_probtraj = None
        
        self.raw_simple_probtraj = None
        self.raw_simple_last_probtraj = None

        self.raw_custom_probtraj = None
        self.raw_custom_last_probtraj = None
        
        self.workdir = workdir

        if workdir is not None and (not os.path.exists(os.path.join(workdir, "%s_run.txt" % prefix)) or overwrite):

            self.cmaboss_result.display_run(os.path.join(workdir, "%s_run.txt" % prefix), hexfloat)
            self.cmaboss_result.display_probtraj(
                os.path.join(workdir, "%s_pop_probtraj.csv" % prefix), 
                os.path.join(workdir, "%s_simple_pop_probtraj.csv" % prefix),
                hexfloat
            )
            self.cmaboss_result.display_fp(os.path.join(workdir, "%s_fp.csv" % prefix), hexfloat)
            # self.cmaboss_result.display_statdist(os.path.join(workdir, "%s_statdist.csv" % prefix))

    def get_raw_last_states_probtraj(self):
        
        if self.raw_last_states_probtraj is None:
            self.raw_last_states_probtraj = self.cmaboss_result.get_last_probtraj()
        return self.raw_last_states_probtraj
    
    def get_raw_states_probtraj(self):
        if self.raw_states_probtraj is None:
            self.raw_states_probtraj = self.cmaboss_result.get_probtraj()
        return self.raw_states_probtraj

    ########### Simple Last Probtraj

    def get_raw_simple_last_probtraj(self):
        if self.raw_simple_last_probtraj is None:
            self.raw_simple_last_probtraj = self.cmaboss_result.get_simple_last_probtraj()
        return self.raw_simple_last_probtraj

    ########### Simple Probtraj
    
    def get_raw_simple_probtraj(self):
        if self.raw_simple_probtraj is None:
            self.raw_simple_probtraj = self.cmaboss_result.get_simple_probtraj()
        return self.raw_simple_probtraj
        
    def get_raw_custom_probtraj(self):
        if self.raw_custom_probtraj is None:
            self.raw_custom_probtraj = self.cmaboss_result.get_custom_probtraj()
        return self.raw_custom_probtraj
        
    def get_raw_custom_last_probtraj(self):
        if self.raw_custom_last_probtraj is None:
            self.raw_custom_last_probtraj = self.cmaboss_result.get_custom_last_probtraj()
        return self.raw_custom_last_probtraj



"""
atexit.py - allow programmer to define multiple exit functions to be executed
upon normal program termination.

One public function, register, is defined.
"""

from __future__ import print_function

__all__ = ["register"]

import sys

_exithandlers = []
def _run_exitfuncs():
    """run any registered exit functions

    _exithandlers is traversed in reverse order so functions are executed
    last in, first out.
    """

    exc_info = None
    while _exithandlers:
        func, targs, kargs = _exithandlers.pop()
        try:
            func(*targs, **kargs)
        except SystemExit:
            exc_info = sys.exc_info()
        except:
            import traceback
            print("Error in atexit._run_exitfuncs:", file=sys.stderr)
            traceback.print_exc()
            exc_info = sys.exc_info()

    if exc_info is not None:
        raise exc_info[0](exc_info[1], exc_info[2])


def register(func, *targs, **kargs):
    """register a function to be executed upon normal program termination

    func - function to be called at exit
    targs - optional arguments to pass to func
    kargs - optional keyword arguments to pass to func

    func is returned to facilitate usage as a decorator.
    """
    _exithandlers.append((func, targs, kargs))
    return func

if hasattr(sys, "exitfunc"):
    # Assume it's another registered exit function - append it to our list
    register(sys.exitfunc)
sys.exitfunc = _run_exitfuncs
